"""Helper functions for performing small tasks using the Ensembl REST API
"""
# import pprint as pp
CHR_NAME = "chr_name"
START_POS = "start_pos"
END_POS = "end_pos"

_BIOTYPE_CACHE = {}
_CHROMOSOME_CACHE = {}


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_ld(rc, chr_name, start_pos, ref_allele, alt_allele, species='human',
           population_name='1000GENOMES:phase_3:EUR', **kwargs):
    """Get variants in LD with the variant coords and alleles, within a
    certain region size.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        The rest client object.
    chr_name : `str`
        The chromosome name for the variant.
    start_pos : `int`
        The start position for the variant.
    ref_allele : `str`
        The reference allele for the variant.
    alt_allele : `str`
        The alternate allele for the variant.
    species : `str`, optional, default: `human`
        The species for the LD panel.
    species : `str`, optional, default: `1000GENOMES:phase_3:EUR`
        The population for the LD panel, default is 1000 genomes European.
    **kwargs
        Keyword aruments to `ensembl_rest_client.Rest.get_ld`. Note, the
        ``attrib`` argument is ignored and on by default.

    Returns
    -------
    ld_matches : `list` of `dict`
        Pairwise LD matches for the query variant.

    Raises
    ------
    KeyError
        If the chromosome coordinates are not associated with an rsID
    """
    # Always get the attribs
    kwargs['attribs'] = True

    # Convert the coordinate positions to an rsID, this will fail with a
    # KeyError if not possible
    var_id = get_var_id(rc, chr_name, start_pos, ref_allele, alt_allele)
    # print(var_id)
    # Get the LD data for the variant
    ld_matches = rc.get_ld(var_id['id'], population_name, species,
                           **kwargs)

    # The variant ID post data size
    post_size = 200

    var_info = {}
    # Because the LD data is not associated with any alleles, we have to use
    # the variant identifiers to get the alleles for the variants
    ids = [i['variation'] for i in ld_matches]
    for i in range(0, len(ids), post_size):
        var_info = {
            **var_info, **(rc.post_variation(species, ids[i:i+post_size]))
        }

    # Now we go through the LD variants and match up the alleles from the
    # variants that we have queried
    for i in ld_matches:
        i['source_variant'] = var_id
        try:
            for j in var_info[i['variation']]['mappings']:
                j['alleles'] = j['allele_string'].split('/')
                if j['seq_region_name'] == i['chr'] \
                   and j['start'] == i['start']:
                    i['ld_variant'] = j
                    break
        except KeyError:
            i['ld_variant'] = {}

    return ld_matches


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_var_id(rc, chr_name, start_pos, ref_allele, alt_allele,
               species='human'):
    """Get a variant rs ID from positional coords and alleles.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        The rest client object.
    chr_name : `str`
        The chromosome name for the variant.
    start_pos : `int`
        The start position for the variant.
    ref_allele : `str`
        The reference allele for the variant.
    alt_allele : `str`
        The alternate allele for the variant.
    species : `str`, optional, default: `human`
        The species for the LD panel.
    species : `str`, optional, default: `1000GENOMES:phase_3:EUR`
        The population for the LD panel, default is 1000 genomes European.
    **kwargs
        Keyword aruments to `ensembl_rest_client.Rest.get_ld`. Note, the
        ``attrib`` argument is ignored and on by default.

    Returns
    -------
    variant : `dict`
        A dictionary containing the data for the matching variant.

    Raises
    ------
    KeyError
        If the chromosome coordinates are not associated with an rsID.
    ValueError
        If the ref allele is the same as the alt allele.

    Notes
    -----
    Note that alleles are matched in a non-ref/alt way, so only ref/alt needs
    to be present in any order for the match to be valid.
    """
    if ref_allele == alt_allele:
        raise ValueError("alleles must be different")

    start_pos, end_pos, ref, alts = vcf_to_ensembl(
        start_pos, ref_allele, [alt_allele], dbcoords=False
    )
    alt_allele = alt_allele[0]

    regions = rc.get_overlap_region(
        'variation', "{0}:{1}-{2}".format(chr_name, start_pos, end_pos),
        species
    )
    for i in regions:
        if ref_allele in i['alleles'] and alt_allele in i['alleles']:
            return i

    raise KeyError("No variant for {0}:{1}|{2}|{3}".format(
        chr_name, start_pos, ref_allele, alt_allele)
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def vcf_to_ensembl(pos, ref, alts, dbcoords=True):
    """Convert VCF coordinates and alleles to ensembl ones, this assumes that
    everything is on the + strand.

    Parameters
    ----------
    pos : `int`
        The VCF POS column
    ref : `str`
        The VCF REF column
    alts : `list` of `str`
        The VCF alternate alleles column
    dbcoords : `bool`, optional, default: `True`
        Adjust INDEL coordinates to match the Ensembl database. In the case of
        INSERTIONS this means start>end. If False then start == end (as in the
        Ensembl GVF files).

    Returns
    -------
    start_pos : `int`
        The ensemblised start position.
    end_pos : `int`
        The ensemblised end position.
    ref_allele : `str`
        The ensemblised reference allele.
    alt_alleles : `list` of `str`
        The ensemblised alternate alleles.
    """
    blank_allele = '-'

    # Make a combined allele list for ease
    alleles = [ref] + alts

    # Make sure that pos is an integer
    pos = int(pos)

    # first make sure they are all DNA_REGEX not that this just tests for
    # ATCG or atcg deletions (-) are not in this regexp. The presence of a -
    # would be indicative of this already being an Ensembl allele
    # is_dna = [
    #     True if constants.DNA_REGEX_DEL.match(i) else False for i in alleles
    # ]
    # # if False in is_dna is True:
    # if False in is_dna:
    #     raise ensembl_errors.SequenceError(
    #         "expected DNA alleles not '{0}'".format("/".join(alleles)))

    # Now make sure it is not already in ensembl format
    is_ensembl = [True if blank_allele == i else False for i in alleles]
    if True in is_ensembl:
        raise ValueError(
            "already ensembl format '{0}'".format("/".join(alleles)))

    # Now get the lengths of the alleles, we only need to do the conversion
    # if the alleles are unbalenced (i.e. not of equal lengths)
    lengths = [len(i) for i in alleles]
    equal_len = [True if len(i) == lengths[0] else False for i in alleles]
    if all(equal_len) is True:
        # In this case we return the start/end and alleles
        # the alleles could still be balenced and > length 1 hence we take
        # the length into account for the end pos
        return pos, pos+len(alleles[0])-1, ref, alts

    # If we get here then we are unbalenced and not in Ensembl format. So in
    # a simple VCF entry we remove the leading base from all the alleles.
    # However, we could have a more complex entry where the leading base is
    # not the same in all scenarios, in these cases we make no allele
    # alterations, this mimicks the behaviour of the VEP

    # This gives the smallest length for an unbalenced VCF allele this should
    # be 1 in the simple, i.e. the common base before the variation. However,
    # Insertion/Deletions and Deletion/Insertions will be different
    smallest = min(lengths)

    # Get the index (i.e. position in the alleles of the smallest one. There is
    # a possibility of a tie for the smallest, in this case we can't convert as
    # both will start with different alleles
    smallest = [c for c, i in enumerate(lengths) if i == smallest]

    # Now we make sure that all the alleles start with te smallest allele and
    # if so we chop it off and make the smallest into a -. Howver, we actually
    # want to check the first base of the smallest allele as there are some
    # deletion insertions. i.e. in VCF TAAAA/TC. i.e. AAAA has been deleted
    # and C has been inserted in it's place
    min_smallest = min(smallest)

    # Get the sequence of the load base so we can test that it is leading in
    # all alleles
    # TODO: Do I need to account for the lead base being . or None?
    # TODO: No I don't as these will raise SequenceErrors above
    lead_base = alleles[min_smallest][0]

    # Loop through the alleles removing the leading bp  if they are present.
    # the modified allele strings are present in new_alleles
    new_alleles = []
    for i in alleles:
        if i.startswith(lead_base):
            ensembl_allele = i[1:]

            # This might make the string empty if so insert the blank allele
            if ensembl_allele == '':
                ensembl_allele = blank_allele

            new_alleles.append(ensembl_allele)
        else:
            # TODO: This is really here for debugging purposes, when I see
            # TODO: examples of these I will adjust
            # If at least one of the alleles does not start with the lead base
            # then we have a mixed variant type. I have noticed that some of
            # the HGMD variants are poorly defined
            # TODO: This will change and is for debugging the rule set
            # I have seen this
            if alleles[1] is None and len(alleles[0]) == 1:
                raise ValueError("Can't convert orphan alleles")
            # print(pos, file=sys.stderr)
            # print(pp.pformat(is_dna), file=sys.stderr)
            # print(False in is_dna, file=sys.stderr)
            # print(pp.pformat(alleles), file=sys.stderr)
            # print(alleles[1], file=sys.stderr)
            # print(pp.pformat(lengths), file=sys.stderr)
            raise RuntimeError("Can't convert mixed variation types")

    # If we get here, then everything has the same lead base
    # Now deal with the coordinates. The rules seem to be
    # 1. Balenced polymorphism then everything stays the same and
    #    END = START + len(REF) - 1
    # 2. Simple Insertion - START = POS + 1; END = POS: so END is > START, this
    #    behaviour can be changed with the dbcoords=False flag
    # 3. Simple Deletion - START = POS+1; END = START + len(REF) -1
    # 4. Complex Insertion - START = POS + 1; END = POS, the same as (2)
    # 5. Complex Deletion - Not observed yet
    # 6. Unbalenced alleles with different first base - Not observed yet so
    #    should stay the same but are errored out above
    START = pos
    END = pos

    # INSERTION i.e. the shortest of the unbaleced alleles is the first
    # index (ref)
    if min_smallest == 0:
        # If we have a simple insertion, i.e. the lead allele is a -
        if new_alleles[0] == blank_allele:
            # if dbcords is True then START > END otherwise START == END
            # START = pos + dbcoords
            # END = pos
            START = pos + 1
            END = pos + (1 * (not dbcoords))
        else:
            # DELETION/INSERTION, i.e. the REF is a longer sequence string.
            # i.e. a base is removed from the REF and a longer sequence is
            # inserted in
            # TODO: Need to check this in a GVF format
            START = pos + 1
            END = START + len(new_alleles[0]) - 1
    else:
        # DELETION
        START = pos + 1
        END = START + len(new_alleles[0]) - 1

    return START, END, new_alleles[0], new_alleles[1:]


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_exons(rc, ensembl_id):
    """Get all the exons and UTR regions for an ensembl gene ID.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    ensembl_gene_id : `str`
        An ensembl gene identifier to get the exons for

    Returns
    -------
    exons : `dict`
        Gene information including exons and UTRs.
    """
    return rc.get_lookup_id(ensembl_id, expand=True, utr=True)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_genes_within(rc, species, chr_name, start_pos, end_pos=None,
                     use_center=False, ref_point_start=None,
                     ref_point_end=None, biotypes=None, use_trans=False):
    """Get the genes that overlap a region +/- n optional flank on top.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    species : `str`
        An Ensembl species name, this is used to get the upper limit of the
        chromosome coordinates.
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`, optional, default: `NoneType`
        The end position of the region of interest. If it is not defined then
        ``end_pos`` = ``start_pos``.
    use_center : `bool`, optional, default: `False`
        Use the center position of the region of interest to define what is
        nearest. If the center point is not an integer it is cast into an
        integer - this will result in rounding down.
    ref_point_start : `int` or `NoneType`, optional, default: `NoneType`
        A reference start point within the selected regions to determine what
        the nearest gene is, if not given then ``start_pos`` is used unless
        ``use_center`` is ``True`` in which case the center point of the region
        is used.
    ref_point_end : `int` or `NoneType`, optional, default: `NoneType`
        A reference end point within the selected regions to determine what the
        nearest gene is, if not given then ``end_pos`` is used unless
        ``use_center`` is ``True`` in which case the center point of the region
        is used.
    biotypes : `NoneType` or `list` of `str`, optional, default: `NoneType`
        Restrict the assessment of the nearest gene to genes with these
        biotypes. ``NoneType`` means no restriction on biotype.
    use_trans : `bool`, optional, default: `False`
        Use the coordinates of the cannonical transcript for each gene, rather
        than the gene coordinates.

    Returns
    -------
    overlap_genes : `dict` or `NoneType`
        Information on the nearest gene (or canonical transcript if
        ``use_trans=True``).

    Notes
    -----
    By default the nearest gene is assess as the gene that is nearest either
    the start_pos or end_pos of the region of interest. The center position can
    be used instead if ``use_center`` is ``True``. Strand is not considered
    when determining the nearest gene. The algorithm works as follows:

    *. Query the length of the chromosome and make sure the region is contained
       within the bounds.
    *. Do any genes already overlap the region, if so then they are all
       returned.
    *. If not, then a region 100kbp upstream and downstream from the region of
       interest is queried and any nearest genes determined.
    *. If none are present regions regions are queried until a nearest gene is
       found or at the ends of the chromosome.
    """
    no_region = (None, None)

    query_func = _query_gene_region
    if use_trans is True:
        query_func = _query_trans_region

    query_window = 100000

    # Process the arguments, this makes sure they are valid and will adjust
    # the start/end if we want to use the center point as a reference
    chr_length, center_point, start_pos, end_pos = _process_args(
        rc, species, chr_name, start_pos, end_pos, use_center, biotypes
    )

    # If the user has not supplied a specific reference point to evaluate
    # against then we use the start_pos/end_pos which could be the center point
    # is use_center=True
    if ref_point_start is None:
        ref_point_start = start_pos
    if ref_point_end is None:
        ref_point_end = end_pos

    seen_genes = set()
    genes = []
    passed_biotype = []
    for left_region, right_region in yield_chr_coords(
            rc, species, chr_name, center_point, center_point,
            batch_length=query_window, left_bounds=start_pos,
            right_bounds=end_pos
    ):
        if left_region != no_region:
            start, end = left_region
            genes.extend(
                query_func(rc, species, chr_name, start, end,
                           seen=seen_genes)
            )
        if right_region != no_region:
            start, end = right_region
            genes.extend(
                query_func(rc, species, chr_name, start, end,
                           seen=seen_genes)
            )

    # If no genes have been found
    if len(genes) == 0:
        return genes

    if biotypes is not None:
        passed_biotype = \
            [i for i in genes if i['biotype'] in biotypes]
    else:
        passed_biotype = genes

    requested_region_length = end_pos - start_pos + 1

    # Calculate the min distance and distance rank according to the reference
    # point
    get_min_dist(chr_name, ref_point_start, ref_point_end, genes)
    for idx, i in enumerate(passed_biotype, 1):
        i['distance_rank_biotype'] = idx
        i['requested_region_length'] = requested_region_length
    return passed_biotype


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_n_nearest_genes(rc, species, chr_name, start_pos, end_pos=None,
                        use_center=False, biotypes=None, ngenes=1,
                        use_trans=False):
    """Get the N nearest genes to the input coords.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    species : `str`
        An Ensembl species name, this is used to get the upper limit of the
        chromosome coordinates.
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`, optional, default: `NoneType`
        The end position of the region of interest. If it is not defined then
        ``end_pos`` = ``start_pos``.
    use_center : `bool`, optional, default: `False`
        Use the center position of the region of interest to define what is
        nearest. If the center point is not an integer it is cast into an
        integer - this will result in rounding down.
    biotypes : `NoneType` or `list` of `str`, optional, default: `NoneType`
        Restrict the assessment of the nearest gene to genes with these
        biotypes. ``NoneType`` means no restriction on biotype.
    ngenes : `int`, optional, default: `1`
        The number of nearest genes to return.
    use_trans : `bool`, optional, default: `False`
        Use the coordinates of the cannonical transcript for each gene, rather
        than the gene coordinates.

    Returns
    -------
    nearest_genes : `dict`
        Information on the nearest gene (or canonical transcript if
        ``use_trans=True``).

    Notes
    -----
    By default the nearest gene is assess as the gene that is nearest either
    the start_pos or end_pos of the region of interest. The center position can
    be used instead if ``use_center`` is ``True``. Strand is not considered
    when determining the nearest gene. The algorithm works as follows:

    *. Query the length of the chromosome and make sure the region is contained
       within the bounds.
    *. Do any genes already overlap the region, if so then they are all
       returned.
    *. If not, then a region 100kbp upstream and downstream from the region of
       interest is queried and any nearest genes determined.
    *. If none are present regions regions are queried until a nearest gene is
       found or at the ends of the chromosome.
    """
    no_region = (None, None)
    query_window = 100000

    query_func = _query_gene_region
    if use_trans is True:
        query_func = _query_trans_region

    # Process the arguments, this makes sure they are valid and will adjust
    # the start/end if we want to use the center point as a reference
    chr_length, center_point, start_pos, end_pos = _process_args(
        rc, species, chr_name, start_pos, end_pos, use_center, biotypes
    )

    seen_genes = set()
    genes = []
    passed_biotype = []
    for left_region, right_region in yield_chr_coords(
            rc, species, chr_name, center_point, center_point,
            batch_length=query_window
    ):
        # print(left_region, right_region)
        if left_region != no_region:
            start, end = left_region
            genes.extend(
                query_func(rc, species, chr_name, start, end,
                           seen=seen_genes)
            )
        if right_region != no_region:
            start, end = right_region
            genes.extend(
                query_func(rc, species, chr_name, start, end,
                           seen=seen_genes)
            )
        if len(genes) >= ngenes:
            if biotypes is not None:
                passed_biotype = \
                    [i for i in genes if i['biotype'] in biotypes]
            else:
                passed_biotype = genes
            if len(passed_biotype) >= ngenes:
                break

    get_min_dist(chr_name, start_pos, end_pos, genes)
    for idx, i in enumerate(passed_biotype, 1):
        i['distance_rank_biotype'] = idx
    return passed_biotype[:ngenes]


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def yield_chr_coords(rc, species, chr_name, start_pos, end_pos,
                     batch_length=100000, left_bounds=1, right_bounds=None):
    """Yield batches for leftwards and rightwards coordinates from a region on
    a chromosome until we reach the extent of the chromosome or the
    ``left_bounds``/``right_bounds``.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    species : `str`
        An Ensembl species name, this is used to get the upper limit of the
        chromosome coordinates.
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`, optional, default: `NoneType`
        The end position of the region of interest. If it is not defined then
        ``end_pos`` = ``start_pos``.
    batch_length : `int`, optional, default: `100000`
        The step range for each yielded left/right coordinates
    left_bounds : `int`, optional, default: `1`
        The lower boundary, when down to this the left coordinates will yield
        (None, None) (assuming the right co-ordinates are still in range).
    right_bounds : `int` or `NoneType`, optional, default: `NoneType`
        The upper boundary, when reached the right coordinates will yield
        (None, None) (assuming the left co-ordinates are still in range).
        ``NoneType`` means the end position of the chromosome.

    Yields
    ------
    left_region : `tuple` of (`int`, `int`) or (`NoneType`, `NoneType`)
        The left region in ``batch_length`` increments. When the left end
        (``left_bounds``) is reached this will be (`NoneType`, `NoneType`).
    right_region : `tuple` of (`int`, `int`) or (`NoneType`, `NoneType`)
        The right region in ``batch_length`` increments. When the right end
        (``right_bounds``) is reached this will be (`NoneType`, `NoneType`).

    Notes
    -----
    This will keep on yielding defined regions of batch_length until
    ``left_bounds`` and ``right_bounds`` is reached. If one or the other
    boundary is reached first we carry on going until the other boundary is
    reached but yield (None, None) for the out of range one.
    """
    try:
        # There is a module level cache of chromosome data to prevent excessive
        # rest queries, so we try to use that first.
        chr_info = _CHROMOSOME_CACHE[species][chr_name]
    except KeyError:
        # Not in the cache, so we query and update the cache
        _CHROMOSOME_CACHE.setdefault(species, {})
        chr_info = rc.get_info_region_assembly(chr_name, species)
        _CHROMOSOME_CACHE[species][chr_name] = chr_info

    # Extract the chromosome length
    chr_length = chr_info['length']

    # NoneType means that we are defaulting tot he max chromosome coordinates
    if right_bounds is None:
        right_bounds = chr_length

    # Ensure any user defined left boundary is within a meaningful range
    left_bounds = max(1, left_bounds)

    # If the user has set the right bounds then make sure it is in range of the
    # chromosome coordinates
    right_bounds = min(right_bounds, chr_length)

    # The left/right coordinate positions that will be updated with each
    # iteration
    left_start = start_pos
    left_end = None
    right_start = None
    right_end = end_pos

    # Booleans that indicate if we want to yield left/right coordinates. When
    # both are False we will exit. If one or the other is False we carry on
    # going but yield (None, None) for the out of range one
    query_left = True
    query_right = True
    while query_left | query_right:
        # The default is out of range
        left_region = (None, None)
        right_region = (None, None)

        # If leftwards is not out of range
        if query_left is True:
            left_end = left_start
            left_start = left_end - batch_length

            # The start position of the left boundary is out of range so we
            # adjust to the limits and indicate that we do not want any more
            # left regions yielded
            if left_start <= left_bounds:
                left_start = left_bounds
                query_left = False
            left_region = (left_start, left_end)

        # If rightwards is not out of range
        if query_right is True:
            right_start = right_end
            right_end = right_start + batch_length

            if right_end >= right_bounds:
                right_end = right_bounds
                query_right = False
            right_region = (right_start, right_end)
        yield left_region, right_region


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def valid_biotypes(rc, species, test_biotypes, object_type=None,
                   use_cache=True):
    """Test to see if biotypes are known valid Ensembl biotypes.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    species : `str`
        An Ensembl species name, this is used to get the upper limit of the
        chromosome coordinates.
    test_biotypes : `list` of `str`
        A list of biotypes to test for validity.
    object_type : `str` or `NoneType`, optional, default: `NoneType`
        The object type for the biotype, so some biotypes are only appropriate
        for genes and others transcripts. Leaving this as ``NoneType`` will
        validate and ignore any object types.
    use_cache : `bool`, optional, default: `True`
        Use the module level cache.

    Raises
    ------
    KeyError
        If any of ``test_biotypes`` are not valid biotypes.

    Notes
    -----
    There is a module level cache of biotypes associated with a species to
    prevent repeated queries to Ensembl. Please note, this is not that
    sophisticated and has no concept of different rest clients. So will not
    known that difference between GRCh37 and GRCh38 biotypes (if there are
    any!?). If this is an issue set ``use_cache=False``.
    """
    try:
        # Will keep a module level cache of the biotypes for the species
        # so we do not have to continually query
        valid_biotypes = _BIOTYPE_CACHE[species]
    except KeyError:
        valid_biotypes = rc.get_info_biotypes(species)

        if use_cache is True:
            _BIOTYPE_CACHE[species] = valid_biotypes

    if object_type is not None:
        valid_biotypes = [
            i['biotype'] for i in valid_biotypes if object_type in i['objects']
        ]
    else:
        valid_biotypes = [
            i['biotype'] for i in valid_biotypes
        ]

    failed_biotypes = [i for i in test_biotypes if i not in valid_biotypes]

    if len(failed_biotypes) > 0:
        raise KeyError(
            "biotypes not valid: '{0}".format(','.join(failed_biotypes))
        )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_min_dist(chr_name, start_pos, end_pos, regions):
    """Calculate the minimum distance between each region and the region
    represented by ``chr_name``, ``start_pos``, ``end_pos``.

    Parameters
    ----------
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`
        The end position of the region of interest.
    regions : `list` of `dict`
        Regions that have been queried via a REST query. Each dict must have
        the keys ``seq_region_name`` (``str``), ``start`` (``int``), ``end``
        (``int``).

    Returns
    -------
    regions : `list` of `dict`
        The regions that were passed to this function but with additional keys
        added. ``abs_min_dist`` (``int``), ``min_dist`` (``int``), ``overlap``
        (``int``), ``distance_rank`` (``int``) and ``region_length`` (``int``).
        All additions are in place.

    Raises
    ------
    ValueError
        If the length of the regions is 0.
    """
    min_start = None
    max_end = None

    if len(regions) == 0:
        raise ValueError("there are no regions to evaluate for min distance")

    for i in regions:
        i['abs_min_dist'] = 5E09
        i['min_dist'] = 5E09
        i['overlap'] = 0

        if i['seq_region_name'] == chr_name:
            # First check for an overlap
            i['overlap'] = get_overlap(
                (chr_name, start_pos, end_pos),
                (i['seq_region_name'], i['start'], i['end'])
            )

            try:
                min_start = min(min_start, i['start'])
                max_end = max(max_end, i['end'])
            except TypeError:
                min_start = i['start']
                max_end = i['end']

            if i['overlap'] > 0:
                i['abs_min_dist'] = -1
                i['min_dist'] = -1
                continue

            distances = sorted(
                [
                    (abs(i), i) for i in [i['start'] - start_pos,
                                          i['end'] - start_pos,
                                          i['start'] - end_pos,
                                          i['end'] - end_pos]
                ],
                key=lambda x: x[0]
            )
            i['abs_min_dist'] = distances[0][0]
            i['min_dist'] = distances[0][1]

    regions.sort(key=lambda x: x['overlap'], reverse=True)
    regions.sort(key=lambda x: x['abs_min_dist'])

    region_length = max_end - min_start + 1
    for idx, i in enumerate(regions, 1):
        i['distance_rank'] = idx
        i['region_length'] = region_length
    return regions


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_region_definition(chr_name, start_pos, end_pos):
    """Generate a text chr/pos scan for use in a rest query.

    Parameters
    ----------
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`
        The end position of the region of interest.

    Returns
    -------
    chr_pos : `str`
        Will have the format ``<chr_name>:<start_pos>-<end_pos>``.

    Notes
    -----
    This does not do any sanity checks.
    """
    return '{0}:{1}-{2}'.format(chr_name, start_pos, end_pos)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_overlap(x, y):
    """Find the overlap between two regions. The start and end coordinates are
    cast to int before comparing.

    Parameters
    ----------
    x : `tuple` (`int`, `int`)
        The min/max of the first feature.
    y : `tuple` (`int`, `int`)
        The min/max of the second feature.

    Returns
    -------
    overlap :`int`
        The overlap between the `x` site and the `y` site or 0 if there is no
        overlap
    """
    if x[0] != y[0]:
        return 0

    # Make sure we have all ints
    # start_x, end_x, start_y, end_y = int(start_x), int(end_x), int(start_y),\
    #      int(end_y)

    # start_y -= 1
    # start_x -= 1

    # The +1 is to get <= or >=
    return max(0, min(x[2], y[2]) - max(x[1], y[1]))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def _process_args(rc, species, chr_name, start_pos, end_pos, use_center,
                  biotypes):
    """Validate arguments given to ``get_genes_within`` and
    ``get_n_nearest_genes``.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    species : `str`
        An Ensembl species name, this is used to get the upper limit of the
        chromosome coordinates.
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`
        The end position of the region of interest. If it is not defined then
        ``end_pos`` = ``start_pos``.
    use_center : `bool`
        Use the center position of the region of interest to define what is
        nearest. If the center point is not an integer it is cast into an
        integer - this will result in rounding down.
    biotypes : `NoneType` or `list` of `str`
        Restrict the assessment of the nearest gene to genes with these
        biotypes. ``NoneType`` means no restriction on biotype.
    """
    end_pos = end_pos or start_pos

    if end_pos < start_pos:
        raise IndexError("end position < start position: {0} < {1}")

    if biotypes is not None:
        # Make sure the biotypes are valid otherwise we could query loads
        valid_biotypes(rc, species, biotypes, object_type='gene')

    # The center point will always be rounded down if it is not an integer
    center_point = int(start_pos + ((end_pos - start_pos)/2))

    if use_center is True:
        start_pos = center_point
        end_pos = start_pos

    try:
        chr_info = _CHROMOSOME_CACHE[species][chr_name]
    except KeyError:
        _CHROMOSOME_CACHE.setdefault(species, {})
        chr_info = rc.get_info_region_assembly(chr_name, species)
        _CHROMOSOME_CACHE[species][chr_name] = chr_info
    chr_length = chr_info['length']

    if start_pos < 1 or end_pos > chr_length:
        raise IndexError(
            "region coordinates out of bounds: chr{0} = {1} start={2},"
            " end={3}".format(chr_name, chr_length, start_pos, end_pos)
        )

    return chr_length, center_point, start_pos, end_pos


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def _query_gene_region(rc, species, chr_name, start_pos, end_pos, seen=None):
    """Query for genes overlapping a region.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    species : `str`
        An Ensembl species name, this is used to get the upper limit of the
        chromosome coordinates.
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`
        The end position of the region of interest
    seen : `set` or `NoneType`, optional, default: `NoneType`
        A set of gene IDs (`str`) that have already been encountered so
        will not be stored and returned.

    Returns
    -------
    all_features : `dict`
        All the genes that overlap the region.
    """
    if seen is None:
        seen = set([])
    feature = 'gene'
    all_features = []

    region_def = get_region_definition(chr_name, start_pos, end_pos)
    for feature in rc.get_overlap_region(feature, region_def, species):
        if feature['id'] not in seen:
            seen.add(feature['id'])
            all_features.append(feature)
    return all_features


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def _query_trans_region(rc, species, chr_name, start_pos, end_pos, seen=None):
    """Query for canonical transcripts overlapping a region.

    Parameters
    ----------
    rc : `ensembl_rest_client.client.Rest`
        A rest client that implements the ``info`` and ``overlap`` interfaces.
    species : `str`
        An Ensembl species name, this is used to get the upper limit of the
        chromosome coordinates.
    chr_name : `str`
        The chromosome that contains the region of interest.
    start_pos : `int`
        The start position of the region of interest.
    end_pos : `int` or `NoneType`
        The end position of the region of interest.
    seen : `set` or `NoneType`, optional, default: `NoneType`
        A set of transcript IDs (`str`) that have already been encountered so
        will not be stored and returned.

    Returns
    -------
    all_features : `dict`
        All the canonical transcripts that overlap the region. Their parent
        gene data is also queried out in a separate query and added as a
        ``dict`` to a ``gene_data`` key.
    """
    if seen is None:
        seen = set([])
    feature = 'transcript'
    all_features = []

    region_def = get_region_definition(chr_name, start_pos, end_pos)
    for feature in rc.get_overlap_region(feature, region_def, species):
        if feature['is_canonical'] == 1 and feature['id'] not in seen:
            seen.add(feature['id'])
            all_features.append(feature)

    # Now query out the gene data for each of the canonical transcripts
    for feature in all_features:
        try:
            feature['gene_data'] = rc.get_lookup_id(
                feature['Parent'], db_type='core', expand=None,
                data_format=None, phenotypes=None, species=species
            )
        except KeyError:
            feature['gene_data'] = dict(
                id=feature['id'], display_name=feature['external_name']
            )

    return all_features



"""Perform queries against the regulation endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Regulation(object):
    """Perform Regulation queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_regulatory_species_microarray_vendor(self, microarray, species,
                                                 vendor,
                                                 norm_func=brc.default_pandas):
        """
        Returns information about a specific microarray. GET method. Queries:
        `regulatory/species/:species/microarray/:microarray/vendor/:vendor <https://rest.ensembl.org/documentation/info/array>`_.

        Parameters
        ----------
        microarray : str
            Microarray name. Example values: 'HumanWG_6_V2'.
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        vendor : str
            Probe name. Example values: 'ILMN_1763508'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = (
            'regulatory/species/{species}/microarray/{microarray}/vendor'
            '/{vendor}'.format(
                species=species,
                microarray=microarray,
                vendor=vendor
            )
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_regulatory_species_epigenome(self, species,
                                         norm_func=brc.default_pandas):
        """
        Returns information about all epigenomes available for the given
        species. GET method. Queries:
        `regulatory/species/:species/epigenome <https://rest.ensembl.org/documentation/info/fetch_all_epigenomes>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'regulatory/species/{species}/epigenome'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_species_binding_matrix(self, binding_matrix, species, unit=None,
                                   norm_func=brc.default_pandas):
        """
        Return the specified binding matrix. GET method. Queries:
        `species/:species/binding_matrix/:binding_matrix_stable_id/ <https://rest.ensembl.org/documentation/info/get_binding_matrix>`_.

        Parameters
        ----------
        binding_matrix : str
            Stable ID of binding matrix. Example values: 'ENSPFM0001'.
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        unit : str or NoneType, optional, default: NoneType
            Unit of the matrix elements. Example values: 'frequencies,
            probabilities, bits'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = (
            'species/{species}/binding_matrix/{binding_matrix}'.format(
                species=species,
                binding_matrix=binding_matrix
            )
        )

        return self.rest_query(
            method_type,
            endpoint,
            unit=unit,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_regulatory_species_microarray(self, species,
                                          norm_func=brc.default_pandas):
        """
        Returns information about all microarrays available for the given
        species. GET method. Queries:
        `regulatory/species/:species/microarray <https://rest.ensembl.org/documentation/info/list_all_microarrays>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'regulatory/species/{species}/microarray'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_regulatory_species_microarray_probe(self, microarray, probe,
                                                species, gene=None,
                                                transcripts=None,
                                                norm_func=brc.default_pandas):
        """
        Returns information about a specific probe from a microarray. GET
        method. Queries:
        `regulatory/species/:species/microarray/:microarray/probe/:probe <https://rest.ensembl.org/documentation/info/probe>`_.

        Parameters
        ----------
        microarray : str
            Microarray name. Example values: 'HumanWG_6_V2'.
        probe : str
            Probe name. Example values: 'ILMN_1763508'.
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        gene : bool or NoneType, optional, default: NoneType
            Has to be used in conjunction with transcript. Displays the
            associated gene.
        transcripts : bool or NoneType, optional, default: NoneType
            Displays the transcripts linked to this probe.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = (
            'regulatory/species/{species}/microarray/{microarray}/'
            'probe/{probe}'.format(
                species=species,
                microarray=microarray,
                probe=probe
            )
        )

        return self.rest_query(
            method_type,
            endpoint,
            gene=gene,
            transcripts=transcripts,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_regulatory_species_microarray_probe_set(self, microarray,
                                                    probe_set, species,
                                                    gene=None,
                                                    transcripts=None,
                                                    norm_func=brc.default_pandas):
        """
        Returns information about a specific probe_set from a microarray. GET
        method. Queries:
        `regulatory/species/:species/microarray/:microarray/probe_set/:probe_set <https://rest.ensembl.org/documentation/info/probe_set>`_.

        Parameters
        ----------
        microarray : str
            Microarray name. Example values: 'HG-U133_Plus_2'.
        probe_set : str
            ProbeSet name. Example values: '202820_at'.
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        gene : bool or NoneType, optional, default: NoneType
            Has to be used in conjunction with transcript. Displays the
            associated gene.
        transcripts : bool or NoneType, optional, default: NoneType
            Displays the transcripts linked to this probe.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = (
            'regulatory/species/{species}/microarray/{microarray}'
            '/probe_set/{probe_set}'.format(
                species=species,
                microarray=microarray,
                probe_set=probe_set
            )
        )

        return self.rest_query(
            method_type,
            endpoint,
            gene=gene,
            transcripts=transcripts,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_regulatory_species_id(self, query_id, species, activity=None,
                                  norm_func=brc.default_pandas):
        """
        Returns a RegulatoryFeature given its stable ID (e.g. ENSR00000082023).
        GET method. Queries:
        `regulatory/species/:species/id/:id <https://rest.ensembl.org/documentation/info/regulatory_id>`_.

        Parameters
        ----------
        query_id : str
            RegulatoryFeature stable ID. Example values: 'ENSR00000082023'.
        species : str
            Species name/alias. Example values: 'homo_sapiens'.
        activity : bool or NoneType, optional, default: NoneType
            Returns the activity of the Regulatory Feature in each Epigenome
            for the given species.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'regulatory/species/{species}/id/{id}'.format(
            species=species,
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            activity=activity,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Regulation(brc._BaseRestClient, _Regulation):
    """Perform Regulation queries
    """
    pass



"""Perform queries against the mapping endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Mapping(object):
    """Perform Mapping queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_map_cdna(self, query_id, region, include_original_region=None,
                     species=None, norm_func=brc.default_pandas):
        """
        Convert from cDNA coordinates to genomic coordinates. Output reflects
        forward orientation coordinates as returned from the Ensembl API. GET
        method. Queries:
        `map/cdna/:id/:region <https://rest.ensembl.org/documentation/info/assembly_cdna>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENST00000288602'.
        region : str
            Query region. Example values: '100..300'.
        include_original_region : bool or NoneType, optional, default: NoneType
            Include original input region (cDNA coordinates) along with the
            target region (genomic coordinates) mappings.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'map/cdna/{id}/{region}'.format(
            id=query_id,
            region=region
        )

        return self.rest_query(
            method_type,
            endpoint,
            include_original_region=include_original_region,
            species=species,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_map_cds(self, query_id, region, include_original_region=None,
                    species=None, norm_func=brc.default_pandas):
        """
        Convert from CDS coordinates to genomic coordinates. Output reflects
        forward orientation coordinates as returned from the Ensembl API. GET
        method. Queries:
        `map/cds/:id/:region <https://rest.ensembl.org/documentation/info/assembly_cds>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENST00000288602'.
        region : str
            Query region. Example values: '1..1000'.
        include_original_region : bool or NoneType, optional, default: NoneType
            Include original input region (cds coordinates) along with the
            target region (genomic coordinates) mappings.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'map/cds/{id}/{region}'.format(
            id=query_id,
            region=region
        )

        return self.rest_query(
            method_type,
            endpoint,
            include_original_region=include_original_region,
            species=species,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_map(self, asm_one, asm_two, region, species, coord_system=None,
                target_coord_system=None, norm_func=brc.default_pandas):
        """
        Convert the co-ordinates of one assembly to another. GET method.
        Queries:
        `map/:species/:asm_one/:region/:asm_two <https://rest.ensembl.org/documentation/info/assembly_map>`_.

        Parameters
        ----------
        asm_one : str
            Version of the input assembly. Example values: 'GRCh37'.
        asm_two : str
            Version of the output assembly. Example values: 'GRCh38'.
        region : str
            Query region. Example values: 'X:1000000..1000100:1
            X:1000000..1000100:-1 X:1000000..1000100'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        coord_system : str or NoneType, optional, default: NoneType
            Name of the input coordinate system. Example values: 'chromosome'.
        target_coord_system : str or NoneType, optional, default: NoneType
            Name of the output coordinate system. Example values: 'chromosome'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'map/{species}/{asm_one}/{region}/{asm_two}'.format(
            species=species,
            asm_one=asm_one,
            region=region,
            asm_two=asm_two
        )

        return self.rest_query(
            method_type,
            endpoint,
            coord_system=coord_system,
            target_coord_system=target_coord_system,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_map_translation(self, query_id, region, species=None,
                            norm_func=brc.default_pandas):
        """
        Convert from protein (translation) coordinates to genomic coordinates.
        Output reflects forward orientation coordinates as returned from the
        Ensembl API. GET method. Queries:
        `map/translation/:id/:region <https://rest.ensembl.org/documentation/info/assembly_translation>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSP00000288602'.
        region : str
            Query region. Example values: '100..300'.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'map/translation/{id}/{region}'.format(
            id=query_id,
            region=region
        )

        return self.rest_query(
            method_type,
            endpoint,
            species=species,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Mapping(brc._BaseRestClient, _Mapping):
    """Perform Mapping queries
    """
    pass



"""Perform queries against the phenotype annotations endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _PhenotypeAnnotations(object):
    """Perform PhenotypeAnnotations queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_phenotype_accession(self, accession, species,
                                include_children=None, include_pubmed_id=None,
                                include_review_status=None, source=None,
                                norm_func=brc.default_pandas):
        """
        Return phenotype annotations for genomic features given a phenotype
        ontology accession. GET method. Queries:
        `/phenotype/accession/:species/:accession <https://rest.ensembl.org/documentation/info/phenotype_accession>`_.

        Parameters
        ----------
        accession : str
            phenotype ontology accession. Example values: 'EFO:0003900'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        include_children : bool or NoneType, optional, default: NoneType
            Include annotations attached to child terms.
        include_pubmed_id : bool or NoneType, optional, default: NoneType
            Include the pubmed_ids.
        include_review_status : bool or NoneType, optional, default: NoneType
            Include the review_status information.
        source : str or NoneType, optional, default: NoneType
            Restrict to annotations from a specific source.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = '/phenotype/accession/{species}/{accession}'.format(
            species=species,
            accession=accession
        )

        return self.rest_query(
            method_type,
            endpoint,
            include_children=include_children,
            include_pubmed_id=include_pubmed_id,
            include_review_status=include_review_status,
            source=source,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_phenotype_gene(self, gene, species, include_associated=None,
                           include_overlap=None, include_pubmed_id=None,
                           include_review_status=None, include_submitter=None,
                           non_specified=None, trait=None, tumour=None,
                           norm_func=brc.default_pandas):
        """
        Return phenotype annotations for a given gene. GET method. Queries:
        `/phenotype/gene/:species/:gene <https://rest.ensembl.org/documentation/info/phenotype_gene>`_.

        Parameters
        ----------
        gene : str
            Query gene name or Ensembl stable ID. Example values:
            'ENSG00000157764'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        include_associated : bool or NoneType, optional, default: NoneType
            Include phenotypes associated with variants reporting this gene.
        include_overlap : bool or NoneType, optional, default: NoneType
            Include phenotypes of features overlapping the gene.
        include_pubmed_id : bool or NoneType, optional, default: NoneType
            Include the pubmed_ids.
        include_review_status : bool or NoneType, optional, default: NoneType
            Include the review_status information.
        include_submitter : bool or NoneType, optional, default: NoneType
            Include the submitter names.
        non_specified : bool or NoneType, optional, default: NoneType
            Return non_specified phenotypes (records that did not provide a
            specific phenotype e.g. 'not provided').
        trait : bool or NoneType, optional, default: NoneType
            Return phenotype/disease associations.
        tumour : bool or NoneType, optional, default: NoneType
            Return mutations observed in tumour samples and the tumour type.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = '/phenotype/gene/{species}/{gene}'.format(
            species=species,
            gene=gene
        )

        return self.rest_query(
            method_type,
            endpoint,
            include_associated=include_associated,
            include_overlap=include_overlap,
            include_pubmed_id=include_pubmed_id,
            include_review_status=include_review_status,
            include_submitter=include_submitter,
            non_specified=non_specified,
            trait=trait,
            tumour=tumour,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_phenotype_region(self, region, species, feature_type=None,
                             include_pubmed_id=None,
                             include_review_status=None,
                             include_submitter=None, non_specified=None,
                             only_phenotypes=None, trait=None, tumour=None,
                             norm_func=brc.default_pandas):
        """
        Return phenotype annotations that overlap a given genomic region. GET
        method. Queries:
        `/phenotype/region/:species/:region <https://rest.ensembl.org/documentation/info/phenotype_region>`_.

        Parameters
        ----------
        region : str
            Query region. A maximum of 5Mb is allowed to be requested at any
            one time. Example values: '9:22125500-22136000:1
            9:22125500-22136000:-1 9:22125500-22136000'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        feature_type : str or NoneType, optional, default: NoneType
            Restrict to phenotype annotations from a specific feature type.
            Example values: 'Variation StructuralVariation Gene QTL'.
        include_pubmed_id : bool or NoneType, optional, default: NoneType
            Include the pubmed_ids.
        include_review_status : bool or NoneType, optional, default: NoneType
            Include the review_status information.
        include_submitter : bool or NoneType, optional, default: NoneType
            Include the submitter names.
        non_specified : bool or NoneType, optional, default: NoneType
            Return non_specified phenotypes (records that did not provide a
            specific phenotype e.g. 'not provided').
        only_phenotypes : bool or NoneType, optional, default: NoneType
            Only returns associated phenotype description and mapped ontology
            accessions for a lighter output.
        trait : bool or NoneType, optional, default: NoneType
            Return phenotype/disease associations.
        tumour : bool or NoneType, optional, default: NoneType
            Return mutations observed in tumour samples and the tumour type.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = '/phenotype/region/{species}/{region}'.format(
            species=species,
            region=region
        )

        return self.rest_query(
            method_type,
            endpoint,
            feature_type=feature_type,
            include_pubmed_id=include_pubmed_id,
            include_review_status=include_review_status,
            include_submitter=include_submitter,
            non_specified=non_specified,
            only_phenotypes=only_phenotypes,
            trait=trait,
            tumour=tumour,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_phenotype_term(self, species, term, include_children=None,
                           include_pubmed_id=None, include_review_status=None,
                           source=None, norm_func=brc.default_pandas):
        """
        Return phenotype annotations for genomic features given a phenotype
        ontology term. GET method. Queries:
        `/phenotype/term/:species/:term <https://rest.ensembl.org/documentation/info/phenotype_term>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        term : str
            phenotype ontology term. Example values: 'coffee consumption'.
        include_children : bool or NoneType, optional, default: NoneType
            Include annotations attached to child terms.
        include_pubmed_id : bool or NoneType, optional, default: NoneType
            Include the pubmed_ids.
        include_review_status : bool or NoneType, optional, default: NoneType
            Include the review_status information.
        source : str or NoneType, optional, default: NoneType
            Restrict to annotations from a specific source.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = '/phenotype/term/{species}/{term}'.format(
            species=species,
            term=term
        )

        return self.rest_query(
            method_type,
            endpoint,
            include_children=include_children,
            include_pubmed_id=include_pubmed_id,
            include_review_status=include_review_status,
            source=source,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class PhenotypeAnnotations(brc._BaseRestClient, _PhenotypeAnnotations):
    """Perform PhenotypeAnnotations queries
    """
    pass



"""
Perform queries against the comparative genomics endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _ComparativeGenomics(object):
    """Perform ComparativeGenomics queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_cafe_genetree_id(self, query_id, compara=None, nh_format=None,
                             norm_func=brc.default_pandas):
        """
        Retrieves a cafe tree of the gene tree using the gene tree stable
        identifier. GET method. Queries:
        `cafe/genetree/id/:id <https://rest.ensembl.org/documentation/info/cafe_tree)>`_`.

        Parameters
        ----------
        query_id : str
            An Ensembl genetree ID. Example values: 'ENSGT00390000003602'.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        nh_format : str or NoneType, optional, default: NoneType
            The format of a NH (New Hampshire) request. Available only with the
            default setting to allow us to return the cafe tree with Taxa names
            appended with number of members and the p_value. example :
            homo_sapiens_3_0.123 where 3 is the number of members and 0.123 is
            the p value.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'cafe/genetree/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            compara=compara,
            nh_format=nh_format,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_cafe_genetree_member_id(self, query_id, compara=None, db_type=None,
                                    nh_format=None, object_type=None,
                                    species=None,
                                    norm_func=brc.default_pandas):
        """
        Retrieves the cafe tree of the gene tree that contains the gene /
        transcript / translation stable identifier. GET method. Queries:
        `cafe/genetree/member/id/:id <https://rest.ensembl.org/documentation/info/cafe_tree_member_id)>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSG00000167664'.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core'.
        nh_format : str or NoneType, optional, default: NoneType
            The format of a NH (New Hampshire) request. Available only with the
            default setting to allow us to return the cafe tree with Taxa names
            appended with number of members and the p_value. example :
            homo_sapiens_3_0.123 where 3 is the number of members and 0.123 is
            the p value.
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene transcript'.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'cafe/genetree/member/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            compara=compara,
            db_type=db_type,
            nh_format=nh_format,
            object_type=object_type,
            species=species,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_cafe_genetree_member_symbol(self, species, symbol, compara=None,
                                        db_type=None, external_db=None,
                                        nh_format=None, object_type=None,
                                        norm_func=brc.default_pandas):
        """
        Retrieves the cafe tree of the gene tree that contains the gene
        identified by a symbol. GET method. Queries:
        `cafe/genetree/member/symbol/:species/:symbol <https://rest.ensembl.org/documentation/info/cafe_tree_member_symbol>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        symbol : str
            Symbol or display name of a gene. Example values: 'BRCA2'.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        external_db : str or NoneType, optional, default: NoneType
            Filter by external database. Example values: 'HGNC'.
        nh_format : str or NoneType, optional, default: NoneType
            The format of a NH (New Hampshire) request. Available only with the
            default setting to allow us to return the cafe tree with Taxa names
            appended with number of members and the p_value. example :
            homo_sapiens_3_0.123 where 3 is the number of members and 0.123 is
            the p value.
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene transcript'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'cafe/genetree/member/symbol/{species}/{symbol}'.format(
            species=species,
            symbol=symbol
        )

        return self.rest_query(
            method_type,
            endpoint,
            compara=compara,
            db_type=db_type,
            external_db=external_db,
            nh_format=nh_format,
            object_type=object_type,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_family_id(self, query_id, aligned=None, compara=None,
                      member_source=None, sequence=None,
                      norm_func=brc.default_pandas):
        """
        Retrieves a family information using the family stable identifier. GET
        method. Queries:
        `family/id/:id <https://rest.ensembl.org/documentation/info/family>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl family ID. Example values: 'PTHR15573'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions). Note: If aligned=1, this will override
            sequence=none.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        member_source : str or NoneType, optional, default: NoneType
            The source of the family members that you want returned. Allowed
            values: all, ensembl, uniprot
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Note: dependant on the setting for
            "aligned", If aligned=1, this will override sequence=none. Allowed
            values: none, cdna, protein
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data

        """
        method_type = 'GET'
        endpoint = 'family/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            compara=compara,
            member_source=member_source,
            sequence=sequence,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_family_member_id(self, query_id, aligned=None, compara=None,
                             member_source=None, sequence=None,
                             norm_func=brc.default_pandas):
        """
        Retrieves the information for all the families that contains the gene /
        transcript / translation stable identifier. GET method. Queries:
        `family/member/id/:id <https://rest.ensembl.org/documentation/info/family_member_id>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSG00000167664'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions). Note: If aligned=1, this will override
            sequence=none.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        member_source : str or NoneType, optional, default: NoneType
            The source of the family members that you want returned. Allowed
            values: all, ensembl, uniprot
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Note: dependant on the setting for
            "aligned", If aligned=1, this will override sequence=none. Allowed
            values: none, cdna, protein
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'family/member/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            compara=compara,
            member_source=member_source,
            sequence=sequence,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_family_member_symbol(self, species, symbol, aligned=None,
                                 compara=None, db_type=None, external_db=None,
                                 member_source=None, object_type=None,
                                 sequence=None, norm_func=brc.default_pandas):
        """
        Retrieves the information for all the families that contains the gene
        identified by a symbol. GET method. Queries:
        `family/member/symbol/:species/:symbol <https://rest.ensembl.org/documentation/info/family_member_symbol>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        symbol : str
            Symbol or display name of a gene. Example values: 'BRCA2'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions). Note: If aligned=1, this will override
            sequence=none.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        external_db : str or NoneType, optional, default: NoneType
            Filter by external database. Example values: 'HGNC'.
        member_source : str or NoneType, optional, default: NoneType
            The source of the family members that you want returned. Allowed
            values: all, ensembl, uniprot
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene transcript'.
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Note: dependant on the setting for
            "aligned", If aligned=1, this will override sequence=none. Allowed
            values: none, cdna, protein
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'family/member/symbol/{species}/{symbol}'.format(
            species=species,
            symbol=symbol
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            compara=compara,
            db_type=db_type,
            external_db=external_db,
            member_source=member_source,
            object_type=object_type,
            sequence=sequence,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_genetree_id(self, query_id, aligned=None, cigar_line=None,
                        clusterset_id=None, compara=None, nh_format=None,
                        prune_species=None, prune_taxon=None, sequence=None,
                        norm_func=brc.default_pandas):
        """
        Retrieves a gene tree for a gene tree stable identifier. GET method.
        Queries:
        `genetree/id/:id <https://rest.ensembl.org/documentation/info/genetree>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl genetree ID. Example values: 'ENSGT00390000003602'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions).
        cigar_line : bool or NoneType, optional, default: NoneType
            Return the aligned sequence encoded in CIGAR format.
        clusterset_id : str or NoneType, optional, default: NoneType
            Name of the gene-tree resource being queried. Common values are
            "default" for the standard multi-clade trees (which exclude all
            non-reference strains) and "murinae" for the trees spanning all
            mouse strains. By default, the most inclusive analysis will be
            selected. Example values: 'default murinae'.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        nh_format : str or NoneType, optional, default: NoneType
            The format of a NH (New Hampshire) request. Allowed values: full,
            display_label_composite, simple, species, species_short_name,
            ncbi_taxon, ncbi_name, njtree, phylip
        prune_species : str or NoneType, optional, default: NoneType
            Prune the tree by species. Supports all species aliases. Will
            return a tree with only the species given. Example values: 'human
            cow'.
        prune_taxon : int or NoneType, optional, default: NoneType
            Prune the tree by taxon. Will return a tree with only the taxons
            given. Example values: '9606 10090'.
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Allowed values: none, cdna, protein
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'genetree/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            cigar_line=cigar_line,
            clusterset_id=clusterset_id,
            compara=compara,
            nh_format=nh_format,
            prune_species=prune_species,
            prune_taxon=prune_taxon,
            sequence=sequence,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_genetree_member_id(self, query_id, aligned=None, cigar_line=None,
                               clusterset_id=None, compara=None, db_type=None,
                               nh_format=None, object_type=None,
                               prune_species=None, prune_taxon=None,
                               sequence=None, species=None,
                               norm_func=brc.default_pandas):
        """
        Retrieves the gene tree that contains the gene / transcript /
        translation stable identifier. GET method. Queries:
        `genetree/member/id/:id <https://rest.ensembl.org/documentation/info/genetree_member_id>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSG00000167664'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions).
        cigar_line : bool or NoneType, optional, default: NoneType
            Return the aligned sequence encoded in CIGAR format.
        clusterset_id : str or NoneType, optional, default: NoneType
            Name of the gene-tree resource being queried. Common values are
            "default" for the standard multi-clade trees (which exclude all
            non-reference strains) and "murinae" for the trees spanning all
            mouse strains. By default, the most inclusive analysis will be
            selected. Example values: 'default murinae'.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core'.
        nh_format : str or NoneType, optional, default: NoneType
            The format of a NH (New Hampshire) request. Allowed values: full,
            display_label_composite, simple, species, species_short_name,
            ncbi_taxon, ncbi_name, njtree, phylip
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene transcript'.
        prune_species : str or NoneType, optional, default: NoneType
            Prune the tree by species. Supports all species aliases. Will
            return a tree with only the species given. Example values: 'human
            cow'.
        prune_taxon : int or NoneType, optional, default: NoneType
            Prune the tree by taxon. Will return a tree with only the taxons
            given. Example values: '9606 10090'.
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Allowed values: none, cdna, protein
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'genetree/member/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            cigar_line=cigar_line,
            clusterset_id=clusterset_id,
            compara=compara,
            db_type=db_type,
            nh_format=nh_format,
            object_type=object_type,
            prune_species=prune_species,
            prune_taxon=prune_taxon,
            sequence=sequence,
            species=species,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_genetree_member_symbol(self, species, symbol, aligned=None,
                                   cigar_line=None, clusterset_id=None,
                                   compara=None, db_type=None,
                                   external_db=None, nh_format=None,
                                   object_type=None, prune_species=None,
                                   prune_taxon=None, sequence=None,
                                   norm_func=brc.default_pandas):
        """
        Retrieves the gene tree that contains the gene identified by a symbol.
        GET method. Queries:
        `genetree/member/symbol/:species/:symbol <https://rest.ensembl.org/documentation/info/genetree_member_symbol>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        symbol : str
            Symbol or display name of a gene. Example values: 'BRCA2'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions).
        cigar_line : bool or NoneType, optional, default: NoneType
            Return the aligned sequence encoded in CIGAR format.
        clusterset_id : str or NoneType, optional, default: NoneType
            Name of the gene-tree resource being queried. Common values are
            "default" for the standard multi-clade trees (which exclude all
            non-reference strains) and "murinae" for the trees spanning all
            mouse strains. By default, the most inclusive analysis will be
            selected. Example values: 'default murinae'.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        external_db : str or NoneType, optional, default: NoneType
            Filter by external database. Example values: 'HGNC'.
        nh_format : str or NoneType, optional, default: NoneType
            The format of a NH (New Hampshire) request. Allowed values: full,
            display_label_composite, simple, species, species_short_name,
            ncbi_taxon, ncbi_name, njtree, phylip
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene transcript'.
        prune_species : str or NoneType, optional, default: NoneType
            Prune the tree by species. Supports all species aliases. Will
            return a tree with only the species given. Example values: 'human
            cow'.
        prune_taxon : int or NoneType, optional, default: NoneType
            Prune the tree by taxon. Will return a tree with only the taxons
            given. Example values: '9606 10090'.
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Allowed values: none, cdna, protein
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'genetree/member/symbol/{species}/{symbol}'.format(
            species=species,
            symbol=symbol
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            cigar_line=cigar_line,
            clusterset_id=clusterset_id,
            compara=compara,
            db_type=db_type,
            external_db=external_db,
            nh_format=nh_format,
            object_type=object_type,
            prune_species=prune_species,
            prune_taxon=prune_taxon,
            sequence=sequence,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_alignment_region(self, region, species, aligned=None, compact=None,
                             compara=None, display_species_set=None, mask=None,
                             method=None, species_set=None,
                             species_set_group=None,
                             norm_func=brc.default_pandas):
        """
        Retrieves genomic alignments as separate blocks based on a region and
        species. GET method. Queries:
        `alignment/region/:species/:region <https://rest.ensembl.org/documentation/info/genomic_alignment_region>`_.

        Parameters
        ----------
        region : str
            Query region. A maximum of 10Mb is allowed to be requested at any
            one time. Example values: 'X:1000000..1000100:1
            X:1000000..1000100:-1 X:1000000..1000100'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions).
        compact : bool or NoneType, optional, default: NoneType
            Applicable to EPO_EXTENDED alignments. If true, concatenate the
            extended species sequences together to create a single sequence.
            Otherwise, separates out all sequences.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        display_species_set : str or NoneType, optional, default: NoneType
            Subset of species in the alignment to be displayed (multiple
            values). All the species in the alignment will be displayed if this
            is not set. Any valid alias may be used. Example values: 'human
            chimp gorilla'.
        mask : str or NoneType, optional, default: NoneType
            Request the sequence masked for repeat sequences. Hard will mask
            all repeats as N's and soft will mask repeats as lowercased
            characters. Example values: 'hard'. Allowed values: hard,soft
        method : str or NoneType, optional, default: NoneType
            The alignment method. Example values: 'PECAN'. Allowed values: epo,
            epo_extended, pecan, lastz_net, blastz_net, translated_blat_net,
            cactus_hal, cactus_hal_pw
        species_set : str or NoneType, optional, default: NoneType
            The set of species used to define the pairwise alignment (multiple
            values). Should not be used with the species_set_group parameter.
            Use /info/compara/species_sets/:method with one of the methods
            listed above to obtain a valid list of species sets. Any valid
            alias may be used. Example values: 'homo_sapiens mus_musculus'.
        species_set_group : str or NoneType, optional, default: NoneType
            The species set group name of the multiple alignment. Should not be
            used with the species_set parameter. Use
            /info/compara/species_sets/:method with one of the methods listed
            above to obtain a valid list of group names. Example values:
            'mammals, amniotes, fish, sauropsids, murinae'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'alignment/region/{species}/{region}'.format(
            species=species,
            region=region
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            compact=compact,
            compara=compara,
            display_species_set=display_species_set,
            mask=mask,
            method=method,
            species_set=species_set,
            species_set_group=species_set_group,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_homology_id(self, query_id, aligned=None, cigar_line=None,
                        compara=None, data_format=None, sequence=None,
                        target_species=None, target_taxon=None,
                        data_type=None, norm_func=brc.default_pandas):
        """
        Retrieves homology information (orthologs) by Ensembl gene id. GET
        method. Queries:
        `homology/id/:id <https://rest.ensembl.org/documentation/info/homology_ensemblgene>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSG00000157764'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions).
        cigar_line : bool or NoneType, optional, default: NoneType
            Return the aligned sequence encoded in CIGAR format.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        data_format : str or NoneType, optional, default: NoneType
            Layout of the response. Allowed values: full, condensed
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Allowed values: none, cdna, protein
        target_species : str or NoneType, optional, default: NoneType
            Filter by species. Supports all species aliases. Example values:
            'human cow'.
        target_taxon : int or NoneType, optional, default: NoneType
            Filter by taxon. Example values: '9606 10090'.
        data_type : str or NoneType, optional, default: NoneType
            The type of homology to return from this call. Projections are
            orthology calls defined between alternative assemblies and the
            genes shared between them. Useful if you need only one type of
            homology back from the service. Allowed values: orthologues,
            paralogues, projections, all
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'homology/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            cigar_line=cigar_line,
            compara=compara,
            format=data_format,
            sequence=sequence,
            target_species=target_species,
            target_taxon=target_taxon,
            type=data_type,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_homology_symbol(self, species, symbol, aligned=None,
                            cigar_line=None, compara=None, external_db=None,
                            data_format=None, sequence=None,
                            target_species=None, target_taxon=None,
                            data_type=None, norm_func=brc.default_pandas):
        """
        Retrieves homology information (orthologs) by symbol. GET method.
        Queries:
        `homology/symbol/:species/:symbol <https://rest.ensembl.org/documentation/info/homology_symbol>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        symbol : str
            Symbol or display name of a gene. Example values: 'BRCA2'.
        aligned : bool or NoneType, optional, default: NoneType
            Return the aligned string if true. Otherwise, return the original
            sequence (no insertions).
        cigar_line : bool or NoneType, optional, default: NoneType
            Return the aligned sequence encoded in CIGAR format.
        compara : str or NoneType, optional, default: NoneType
            Name of the compara database to use. Multiple comparas exist on a
            server for separate species divisions. Example values:
            'vertebrates'.
        external_db : str or NoneType, optional, default: NoneType
            Filter by external database. Example values: 'HGNC'.
        data_format : str or NoneType, optional, default: NoneType
            Layout of the response. Allowed values: full,condensed
        sequence : str or NoneType, optional, default: NoneType
            The type of sequence to bring back. Setting it to none results in
            no sequence being returned. Allowed values: none, cdna, protein
        target_species : str or NoneType, optional, default: NoneType
            Filter by species. Supports all species aliases. Example values:
            'homo_sapiens human'.
        target_taxon : int or NoneType, optional, default: NoneType
            Filter by taxon. Example values: '9606 10090'.
        data_type : str or NoneType, optional, default: NoneType
            The type of homology to return from this call. Projections are
            orthology calls defined between alternative assemblies and the
            genes shared between them. Useful if you need only one type of
            homology back from the service. Allowed values: orthologues,
            paralogues, projections, all
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'homology/symbol/{species}/{symbol}'.format(
            species=species,
            symbol=symbol
        )

        return self.rest_query(
            method_type,
            endpoint,
            aligned=aligned,
            cigar_line=cigar_line,
            compara=compara,
            external_db=external_db,
            format=data_format,
            sequence=sequence,
            target_species=target_species,
            target_taxon=target_taxon,
            type=data_type,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class ComparativeGenomics(brc._BaseRestClient, _ComparativeGenomics):
    """Perform ComparativeGenomics queries
    """
    pass



"""Perform queries against the overlap endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Overlap(object):
    """Perform Overlap queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_overlap_id(self, feature, query_id, biotype=None, db_type=None,
                       logic_name=None, misc_set=None, object_type=None,
                       so_term=None, species=None, species_set=None,
                       variant_set=None):
        """Retrieves features (e.g. genes, transcripts, variants and more) that
        overlap a region defined by the given identifier. GET method. Queries:
        `overlap/id/:id <https://rest.ensembl.org/documentation/info/overlap_id>`_.
        Max slice length: 5e6.

        Parameters
        ----------
        feature : `str`
            The type of feature to retrieve. Multiple values are accepted.
            Allowed values: band, gene, transcript, cds, exon, repeat, simple,
            misc, variation, somatic_variation, structural_variation,
            somatic_structural_variation, constrained, regulatory, motif,
            chipseq, array_probe
        query_id : `str`
            An Ensembl stable ID. Example values: 'ENSG00000157764'.
        biotype : `str` or `NoneType`, optional, default: `NoneType`
            The functional classification of the gene or transcript to fetch.
            Cannot be used in conjunction with logic_name when querying
            transcripts. Example values: 'protein_coding'.
        db_type : `str` or `NoneType`, optional, default: `NoneType`
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core'.
        logic_name : `str` or `NoneType`, optional, default: `NoneType`
            Limit retrieval of genes, transcripts and exons by a given name of
            an analysis.
        misc_set : `str` or `NoneType`, optional, default: `NoneType`
            Miscellaneous set which groups together feature entries. Consult
            the DB or returned data sets to discover what is available. Example
            values: 'cloneset_30k'.
        object_type : `str` or `NoneType`, optional, default: `NoneType`
            Filter by feature type. Example values: 'gene'.
        so_term : `str` or `NoneType`, optional, default: `NoneType`
            Sequence Ontology term to narrow down the possible variants
            returned. Example values: 'SO:0001650'.
        species : `str` or `NoneType`, optional, default: `NoneType`
            Species name/alias. Example values: 'homo_sapiens'.
        species_set : `str` or `NoneType`, optional, default: `NoneType`
            Filter by species set for retrieving constrained elements.
        variant_set : `str` or `NoneType`, optional, default: `NoneType`
            Short name of a set to restrict the variants found. (See list of
            short set names). Example values: 'ClinVar'.
        """
        method_type = 'GET'
        slice_length = 5e6
        endpoint = 'overlap/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            slice_length=slice_length,
            feature=feature,
            biotype=biotype,
            db_type=db_type,
            logic_name=logic_name,
            misc_set=misc_set,
            object_type=object_type,
            so_term=so_term,
            species=species,
            species_set=species_set,
            variant_set=variant_set
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_overlap_region(self, feature, region, species, biotype=None,
                           db_type=None, logic_name=None, misc_set=None,
                           so_term=None, species_set=None,
                           trim_downstream=None, trim_upstream=None,
                           variant_set=None):
        """Retrieves features (e.g. genes, transcripts, variants and more)
        that overlap a given region. GET method. Queries:
        `overlap/region/:species/:region <https://rest.ensembl.org/documentation/info/overlap_region>`_.
        Max slice length: 5e6.

        Parameters
        ----------
        feature : `str`
            The type of feature to retrieve. Multiple values are accepted.
            Allowed values: band, gene, transcript, cds, exon, repeat, simple,
            misc, variation, somatic_variation, structural_variation,
            somatic_structural_variation, constrained, regulatory, motif, peak,
            other_regulatory, array_probe
        region : `str`
            Query region. A maximum of 5Mb is allowed to be requested at any
            one time. Example values: 'X:1..1000:1 X:1..1000:-1 X:1..1000'.
        species : `str`
            Species name/alias. Example values: 'homo_sapiens'.
        biotype : `str` or `NoneType`, optional, default: `NoneType`
            Functional classification of the gene or transcript to fetch.
            Cannot be used in conjunction with logic_name when querying
            transcripts. Example values: 'protein_coding'.
        db_type : `str` or `NoneType`, optional, default: `NoneType`
            Specify the database type to retrieve features from if not using
            the core database. We automatically choose the correct type of DB
            for variation, comparative and regulation features. Example values:
            'core otherfeatures'.
        logic_name : `str` or `NoneType`, optional, default: `NoneType`
            Limit retrieval of genes, transcripts and exons by the name of
            analysis.
        misc_set : `str` or `NoneType`, optional, default: `NoneType`
            Miscellaneous set which groups together feature entries. Consult
            the DB or returned data sets to discover what is available. Example
            values: 'cloneset_30k'.
        so_term : `str` or `NoneType`, optional, default: `NoneType`
            Sequence Ontology term to restrict the variants found. Its
            descendants are also included in the search. Example values:
            'SO:0001650'.
        species_set : `str` or `NoneType`, optional, default: `NoneType`
            The species set name for retrieving constrained elements.
        trim_downstream : `bool` or `NoneType`, optional, default: `NoneType`
            Do not return features which overlap the downstream end of the
            region.
        trim_upstream : `bool` or `NoneType`, optional, default: `NoneType`
            Do not return features which overlap upstream end of the region.
        variant_set : `str` or `NoneType`, optional, default: `NoneType`
            Short name of a set to restrict the variants found. (See list of
            short set names). Example values: 'ClinVar'.
        """
        method_type = 'GET'
        slice_length = 5e6
        endpoint = 'overlap/region/{species}/{region}'.format(
            species=species,
            region=region
        )

        return self.rest_query(
            method_type,
            endpoint,
            slice_length=slice_length,
            feature=feature,
            biotype=biotype,
            db_type=db_type,
            logic_name=logic_name,
            misc_set=misc_set,
            so_term=so_term,
            species_set=species_set,
            trim_downstream=trim_downstream,
            trim_upstream=trim_upstream,
            variant_set=variant_set
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_overlap_translation(self, query_id, db_type=None, feature=None,
                                so_term=None, species=None, data_type=None):
        """Retrieve features related to a specific Translation as described by
        its stable ID (e.g. domains, variants). GET method. Queries:
        `overlap/translation/:id <https://rest.ensembl.org/documentation/info/overlap_translation>`_.
        Max slice length: 5e6.

        Parameters
        ----------
        query_id : `str`
            An Ensembl stable ID. Example values: 'ENSP00000288602'.
        db_type : `str` or `NoneType`, optional, default: `NoneType`
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core'.
        feature : `str` or `NoneType`, optional, default: `NoneType`
            Specify the type of features requested for the translation. Allowed
            values: transcript_variation, protein_feature, residue_overlap,
            translation_exon, somatic_transcript_variation
        so_term : `str` or `NoneType`, optional, default: `NoneType`
            Sequence Ontology term to restrict the variants found. Its
            descendants are also included in the search. Example values:
            'SO:0001650'.
        species : `str` or `NoneType`, optional, default: `NoneType`
            Species name/alias. Example values: 'homo_sapiens'.
        data_type : `str` or `NoneType`, optional, default: `NoneType`
            Type of data to filter by. By default, all features are returned.
            Can specify a domain or consequence type. Example values:
            'low_complexity'.
        """
        method_type = 'GET'
        slice_length = 5e6
        endpoint = 'overlap/translation/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            slice_length=slice_length,
            db_type=db_type,
            feature=feature,
            so_term=so_term,
            species=species,
            type=data_type
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Overlap(brc._BaseRestClient, _Overlap):
    """Perform Overlap queries
    """
    pass



"""A cache for the REST results borrows heavily from the ideas here:
https://github.com/ubershmekel/filecache/tree/master/filecache
"""
from pathlib import Path
from collections import namedtuple
from time import time
import hashlib
import pickle
import os
import codecs
import shelve
import atexit


CACHE_ROOT = os.path.join(os.environ['HOME'], ".cache", "ensembl_cache")
Path(CACHE_ROOT).mkdir(parents=True, exist_ok=True)
CACHE_PATH = os.path.join(CACHE_ROOT, "ensembl.cache")

MINUTE = 60
HOUR = 60 * MINUTE
DAY = 24 * HOUR
MONTH = 30 * DAY
YEAR = 365 * DAY

CacheEntry = namedtuple("CacheEntry", ['time', 'data'])


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class DummyCache(object):
    """A dummy cache that does not do anything but provide an interface.

    Parameters
    ----------
    *args
        Ignored.
    **kwargs
        Ignored.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, **kwargs):
        pass

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __len__(self):
        """Length will always be 0
        """
        return 0

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @staticmethod
    def get_cache_for_url(url):
        """Generate a cache file name based on the input URL.

        Parameters
        ----------
        args : `url`
            The input URL.

        Returns
        -------
        cache_file : `str`
            The cache file for the URL. Filenames are based on an SHA1 checksum
            of the URL.
        """
        # Ensure the URLs all end the same
        if not url.endswith('/'):
            url = r'{url}/'

        d = hashlib.sha1(url.encode())
        file_basename = d.hexdigest()
        return os.path.join(CACHE_PATH, file_basename)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @staticmethod
    def get_key(args):
        """Get a string based on function args/kwargs that that can be used as a
        key for the cache.

        Parameters
        ----------
        args : `tuple`
            The arguments that can be used as a str key for Shelve.
        """
        return codecs.encode(
            pickle.dumps(
                args,
                protocol=0
            ), "base64").decode()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def purge_cache(self, *args, **kwargs):
        """A dummy purge cache method.

        Parameters
        ----------
        *args
            Ignored.
        **kwargs
            Ignored.
        """
        pass

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def add_to_cache(self, *args, **kwargs):
        """A dummy add to cache method.

        Parameters
        ----------
        *args
            Ignored.
        **kwargs
            Ignored.
        """
        pass

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_from_cache(self, *args, **kwargs):
        """A dummy get from cache method.

        Parameters
        ----------
        *args
            Ignored.
        **kwargs
            Ignored.
        """
        return None


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class ShelveCache(DummyCache):
    """Create a file cache using Python's Shelve module.

    Parameters
    ----------
    cache_path : `str`
        The path to the cache file. This may or may not exist.
    seconds_cutoff : `int`, optional, default: `NoneType`
        Only entries < this will be deemed valid cache entries and
        returned. If ``NoneType``, then any matching entry is returned.
    purge : `bool`, optional, default: `True`
        Go through the cache on setup and remove any entries >
        ``seconds_cutoff`` old.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, cache_path, seconds_cutoff=None, purge=True):
        cache_db = shelve.open(cache_path)
        atexit.register(cache_db.close)
        self.cache_db = cache_db
        self.seconds_cutoff = seconds_cutoff

        if purge is True:
            self.purge_cache()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __len__(self):
        """The number of entries in the cache.
        """
        return len(self.cache_db)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def purge_cache(self):
        """Go through the cache on setup and remove any entries >
        ``ShelveCache.seconds_cutoff`` old.

        Returns
        -------
        n_deleted : `int`
            The number of entries purged from the cache.
        """
        to_remove = []
        if self.seconds_cutoff is not None and self.seconds_cutoff > 0:
            cur_time = time()

            for key in self.cache_db.keys():
                entry = self.cache_db[key]
                if cur_time - entry.time >= self.seconds_cutoff:
                    to_remove.append(key)

            for i in to_remove:
                del self.cache_db[i]
        return len(to_remove)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_from_cache(self, key):
        """Get data from the cache (if it exists).

        Parameters
        ----------
        key : `str`
            The key that the data was stored under. This will usualy be the
            result of the ``ShelveCache.get_key`` call.

        Returns
        -------
        data : `Any`
            The data that was stored in the cache. If it is ``NoneType``, then
            no matching data exists in the cache.
        """
        if key in self.cache_db:
            entry = self.cache_db[key]
            if self.seconds_cutoff is None or \
               time() - entry.time < self.seconds_cutoff:
                return entry.data
        return None

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def add_to_cache(self, key, data):
        """Add the data to the cache.

        Parameters
        ----------
        key : `str`
            The key that the data will be stored under. This will usualy be the
            result of the ``ShelveCache.get_key`` call.
        data : `Any`
            The data to add, must be pickle-able.
        """
        self.cache_db[key] = CacheEntry(data=data, time=time())
        self.cache_db.sync()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_cache(cache_path, url=None, **kwargs):
    """Setup the cache depending on the ``cache_path``.

    Parameters
    ----------
    cache_path : `str` or `bool` or `NoneType`
        The path to the cache, `False` or `NoneType` means you do not want to
        implement a cache. `True` means the cache filename will be generated
        from the `url`, A string is interpreted as a file name.
    url : `str`, optional, default: `NoneType`
        If the `cache_path` is set to `True`, then the `url` must be defined.
    **kwargs
        Any keyword arguments for the cache.

    Returns
    -------
    cache : `ShelveCache` or `DummyCache`
        If a string ``cache_path`` is given then a ``ShelveCache`` object is
        returned, if ``NoneType`` then a ``DummyCache`` is returned, this
        implements the exact same interface but does nothing.
    """
    if cache_path is False or cache_path is None:
        return DummyCache(cache_path, **kwargs)

    if cache_path is True:
        if url is None:
            raise TypeError("url must be defined if cache is set to True")
        cache_path = ShelveCache.get_cache_for_url(url)

    return ShelveCache(cache_path, **kwargs)



"""Perform queries against the archive endpoints
"""

from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Archive(object):
    """Perform queries against the archive endpoints
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_archive_id(self, query_id, norm_func=brc.default_pandas):
        """
        Uses the given identifier to return its latest version. GET method.
        Queries:
        `archive/id/:id <https://rest.ensembl.org/documentation/info/archive_id_get>`_

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSG00000157764'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        result_format = 'dict'
        endpoint = 'archive/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_archive_id(self, query_id, norm_func=brc.default_pandas):
        """
        Retrieve the latest version for a set of identifiers. POST method.
        Queries:
        `archive/id <https://rest.ensembl.org/documentation/info/archive_id_post>`_
        Max post size: 1000.

        Parameters
        ----------
        query_id : list
            Post option not defined in required/optional args.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data

        """
        method_type = 'POST'
        result_format = 'list'
        maximum_post_size = 1000
        endpoint = 'archive/id'

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            id=query_id,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Archive(brc._BaseRestClient, _Archive):
    """Perform queries against the archive endpoints
    """
    pass



"""Perform queries against the sequence endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Sequence(object):
    """Perform Sequence queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_sequence_id(self, query_id, db_type=None, end=None,
                        expand_3prime=None, expand_5prime=None,
                        data_format=None, mask=None, mask_feature=None,
                        multiple_sequences=None, object_type=None,
                        species=None, start=None, data_type=None,
                        norm_func=brc.default_pandas):
        """
        Request multiple types of sequence by stable identifier. Supports
        feature masking and expand options. GET method. Queries:
        `sequence/id/:id <https://rest.ensembl.org/documentation/info/sequence_id>`_.
        Max slice length: 1e7.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSG00000157764
            ENSG00000157764.fasta (supported on some deployments)'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core'.
        end : int or NoneType, optional, default: NoneType
            Trim the end of the sequence by this many basepairs. Trimming is
            relative to reading direction and in the coordinate system of the
            stable identifier. Parameter can not be used in conjunction with
            expand_5prime or expand_3prime. Example values: '1000'.
        expand_3prime : int or NoneType, optional, default: NoneType
            Expand the sequence downstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        expand_5prime : int or NoneType, optional, default: NoneType
            Expand the sequence upstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        data_format : str or NoneType, optional, default: NoneType
            Format of the data. Example values: 'fasta'. Allowed values: fasta
        mask : str or NoneType, optional, default: NoneType
            Request the sequence masked for repeat sequences. Hard will mask
            all repeats as N's and soft will mask repeats as lowercased
            characters. Only available when using genomic sequence type.
            Example values: 'hard'. Allowed values: hard,soft
        mask_feature : bool or NoneType, optional, default: NoneType
            Mask features on the sequence. If sequence is genomic, mask
            introns. If sequence is cDNA, mask UTRs. Incompatible with the
            'mask' option.
        multiple_sequences : bool or NoneType, optional, default: NoneType
            Allow the service to return more than 1 sequence per identifier.
            This is useful when querying for a gene but using a type such as
            protein.
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene'.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens'.
        start : int or NoneType, optional, default: NoneType
            Trim the start of the sequence by this many basepairs. Trimming is
            relative to reading direction and in the coordinate system of the
            stable identifier. Parameter can not be used in conjunction with
            expand_5prime or expand_3prime. Example values: '1000'.
        data_type : str or NoneType, optional, default: NoneType
            Type of sequence. Defaults to genomic where applicable, i.e. not
            translations. cdna refers to the spliced transcript sequence with
            UTR; cds refers to the spliced transcript sequence without UTR.
            Example values: 'cds'. Allowed values: genomic,cds,cdna,protein
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        slice_length = 1e7
        endpoint = 'sequence/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            slice_length=slice_length,
            db_type=db_type,
            end=end,
            expand_3prime=expand_3prime,
            expand_5prime=expand_5prime,
            format=data_format,
            mask=mask,
            mask_feature=mask_feature,
            multiple_sequences=multiple_sequences,
            object_type=object_type,
            species=species,
            start=start,
            type=data_type,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_sequence_id(self, ids, db_type=None, end=None,
                         expand_3prime=None, expand_5prime=None,
                         data_format=None, mask=None, mask_feature=None,
                         object_type=None, species=None, start=None,
                         data_type=None, norm_func=brc.default_pandas):
        """
        Request multiple types of sequence by a stable identifier list. POST
        method. Queries:
        `sequence/id <https://rest.ensembl.org/documentation/info/sequence_id_post>`_.
        Max post size: 50. Max slice length: 1e7.

        Parameters
        ----------
        ids : list
            Post option not defined in required/optional args.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core'.
        end : int or NoneType, optional, default: NoneType
            Trim the end of the sequence by this many basepairs. Trimming is
            relative to reading direction and in the coordinate system of the
            stable identifier. Parameter can not be used in conjunction with
            expand_5prime or expand_3prime. Example values: '1000'.
        expand_3prime : int or NoneType, optional, default: NoneType
            Expand the sequence downstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        expand_5prime : int or NoneType, optional, default: NoneType
            Expand the sequence upstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        data_format : str or NoneType, optional, default: NoneType
            Format of the data. Example values: 'fasta'. Allowed values: fasta
        mask : str or NoneType, optional, default: NoneType
            Request the sequence masked for repeat sequences. Hard will mask
            all repeats as N's and soft will mask repeats as lowercased
            characters. Only available when using genomic sequence type.
            Example values: 'hard'. Allowed values: hard,soft
        mask_feature : bool or NoneType, optional, default: NoneType
            Mask features on the sequence. If sequence is genomic, mask
            introns. If sequence is cDNA, mask UTRs. Incompatible with the
            'mask' option.
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene'.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens'.
        start : int or NoneType, optional, default: NoneType
            Trim the start of the sequence by this many basepairs. Trimming is
            relative to reading direction and in the coordinate system of the
            stable identifier. Parameter can not be used in conjunction with
            expand_5prime or expand_3prime. Example values: '1000'.
        data_type : str or NoneType, optional, default: NoneType
            Type of sequence. Defaults to genomic where applicable, i.e. not
            translations. cdna refers to the spliced transcript sequence with
            UTR; cds refers to the spliced transcript sequence without UTR.
            Example values: 'cds'. Allowed values: genomic,cds,cdna,protein
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        maximum_post_size = 50
        slice_length = 1e7
        endpoint = 'sequence/id'

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            slice_length=slice_length,
            ids=ids,
            db_type=db_type,
            end=end,
            expand_3prime=expand_3prime,
            expand_5prime=expand_5prime,
            format=data_format,
            mask=mask,
            mask_feature=mask_feature,
            object_type=object_type,
            species=species,
            start=start,
            type=data_type,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_sequence_region(self, region, species, coord_system=None,
                            coord_system_version=None, expand_3prime=None,
                            expand_5prime=None, data_format=None, mask=None,
                            mask_feature=None, norm_func=brc.default_pandas):
        """
        Returns the genomic sequence of the specified region of the given
        species. Supports feature masking and expand options. GET method.
        Queries:
        `sequence/region/:species/:region <https://rest.ensembl.org/documentation/info/sequence_region>`_.
        Max slice length: 1e7.

        Parameters
        ----------
        region : str
            Query region. A maximum of 10Mb is allowed to be requested at any
            one time. Example values: 'X:1000000..1000100:1
            X:1000000..1000100:-1 X:1000000..1000100'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        coord_system : str or NoneType, optional, default: NoneType
            Filter by coordinate system name. Example values: 'contig
            seqlevel'.
        coord_system_version : str or NoneType, optional, default: NoneType
            Filter by coordinate system version. Example values: 'GRCh37'.
        expand_3prime : int or NoneType, optional, default: NoneType
            Expand the sequence downstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        expand_5prime : int or NoneType, optional, default: NoneType
            Expand the sequence upstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        data_format : str or NoneType, optional, default: NoneType
            Format of the data. Example values: 'fasta'. Allowed values: fasta
        mask : str or NoneType, optional, default: NoneType
            Request the sequence masked for repeat sequences. Hard will mask
            all repeats as N's and soft will mask repeats as lower cased
            characters. Only available when using genomic sequence type.
            Example values: 'hard'. Allowed values: hard,soft
        mask_feature : bool or NoneType, optional, default: NoneType
            Mask features on the sequence. If sequence is genomic, mask
            introns. If sequence is cDNA, mask UTRs. Incompatible with the
            'mask' option. Example values: '1'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        slice_length = 1e7
        endpoint = 'sequence/region/{species}/{region}'.format(
            species=species,
            region=region
        )

        return self.rest_query(
            method_type,
            endpoint,
            slice_length=slice_length,
            coord_system=coord_system,
            coord_system_version=coord_system_version,
            expand_3prime=expand_3prime,
            expand_5prime=expand_5prime,
            format=data_format,
            mask=mask,
            mask_feature=mask_feature,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_sequence_region(self, species, regions, coord_system=None,
                             coord_system_version=None, expand_3prime=None,
                             expand_5prime=None, data_format=None, mask=None,
                             mask_feature=None, norm_func=brc.default_pandas):
        """
        Request multiple types of sequence by a list of regions. POST method.
        Queries:
        `sequence/region/:species <https://rest.ensembl.org/documentation/info/sequence_region_post>`_.
        Max post size: 50. Max slice length: 1e7.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        regions : list
            Post option not defined in required/optional args.
        coord_system : str or NoneType, optional, default: NoneType
            Filter by coordinate system name. Example values: 'contig
            seqlevel'.
        coord_system_version : str or NoneType, optional, default: NoneType
            Filter by coordinate system version. Example values: 'GRCh37'.
        expand_3prime : int or NoneType, optional, default: NoneType
            Expand the sequence downstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        expand_5prime : int or NoneType, optional, default: NoneType
            Expand the sequence upstream of the sequence by this many
            basepairs. Only available when using genomic sequence type. Example
            values: '1000'.
        data_format : str or NoneType, optional, default: NoneType
            Format of the data. Example values: 'fasta'. Allowed values: fasta
        mask : str or NoneType, optional, default: NoneType
            Request the sequence masked for repeat sequences. Hard will mask
            all repeats as N's and soft will mask repeats as lower cased
            characters. Only available when using genomic sequence type.
            Example values: 'hard'. Allowed values: hard,soft
        mask_feature : bool or NoneType, optional, default: NoneType
            Mask features on the sequence. If sequence is genomic, mask
            introns. If sequence is cDNA, mask UTRs. Incompatible with the
            'mask' option. Example values: '1'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        maximum_post_size = 50
        slice_length = 1e7
        endpoint = 'sequence/region/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            slice_length=slice_length,
            regions=regions,
            coord_system=coord_system,
            coord_system_version=coord_system_version,
            expand_3prime=expand_3prime,
            expand_5prime=expand_5prime,
            format=data_format,
            mask=mask,
            mask_feature=mask_feature,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Sequence(brc._BaseRestClient, _Sequence):
    """Perform Sequence queries
    """
    pass



"""Perform queries against the linkage disequilibrium endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _LinkageDisequilibrium(object):
    """Perform LinkageDisequilibrium queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ld(self, query_id, population_name, species, attribs=None,
               d_prime=None, r2=None, window_size=None,
               norm_func=brc.default_pandas):
        """
        Computes and returns LD values between the given variant and all other
        variants in a window centered around the given variant. The window size
        is set to 500 kb. GET method. Queries:
        `ld/:species/:id/:population_name <https://rest.ensembl.org/documentation/info/ld_id_get>`_.

        Parameters
        ----------
        query_id : str
            Variant id. Example values: 'rs56116432'.
        population_name : str
            Population for which to compute LD. Use GET
            /info/variation/populations/:species?filter=LD to retrieve a list
            of all populations with LD data. Example values:
            '1000GENOMES:phase_3:KHV'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        attribs : bool or NoneType, optional, default: NoneType
            Add variation attributes for the variation which is used to compute
            LD data with the input variation: chr, start, end, strand,
            consequence_type, clinical_significance.
        d_prime : float or NoneType, optional, default: NoneType
            Measure of LD. If D' is provided only return pairs of variants
            whose D' value is equal to or greater than the value provided.
            Example values: '1.0'.
        r2 : float or NoneType, optional, default: NoneType
            Measure of LD. If r-squared is provided only return pairs of
            variants whose r-squared value is equal to or greater than the
            value provided. Example values: '0.85'.
        window_size : int or NoneType, optional, default: NoneType
            Window size in kb. The maximum allowed value for the window size is
            500 kb. LD is computed for the given variant and all variants that
            are located within the specified window. Example values: '500'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ld/{species}/{id}/{population_name}'.format(
            species=species,
            id=query_id,
            population_name=population_name
        )

        return self.rest_query(
            method_type,
            endpoint,
            attribs=attribs,
            d_prime=d_prime,
            r2=r2,
            window_size=window_size
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ld_pairwise(self, id1, id2, species, d_prime=None,
                        population_name=None, r2=None,
                        norm_func=brc.default_pandas):
        """
        Computes and returns LD values between the given variants. GET method.
        Queries:
        `ld/:species/pairwise/:id1/:id2 <https://rest.ensembl.org/documentation/info/ld_pairwise_get>`_.

        Parameters
        ----------
        id1 : str
            Variant id1. Example values: 'rs6792369'.
        id2 : str
            Variant id2. Example values: 'rs1042779'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        d_prime : float or NoneType, optional, default: NoneType
            Measure of LD. If D' is provided only return pairs of variants
            whose D' value is equal to or greater than the value provided.
            Example values: '1.0'.
        population_name : str or NoneType, optional, default: NoneType
            Only compute LD for this population. Use GET
            /info/variation/populations/:species?filter=LD to retrieve a list
            of all populations with LD data. Example values:
            '1000GENOMES:phase_3:KHV'.
        r2 : float or NoneType, optional, default: NoneType
            Measure of LD. If r-squared is provided only return pairs of
            variants whose r-squared value is equal to or greater than the
            value provided. Example values: '0.85'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ld/{species}/pairwise/{id1}/{id2}'.format(
            species=species,
            id1=id1,
            id2=id2
        )

        return self.rest_query(
            method_type,
            endpoint,
            d_prime=d_prime,
            population_name=population_name,
            r2=r2
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ld_region(self, population_name, region, species, d_prime=None,
                      r2=None, norm_func=brc.default_pandas):
        """
        Computes and returns LD values between all pairs of variants in the
        defined region. GET method. Queries:
        `ld/:species/region/:region/:population_name <https://rest.ensembl.org/documentation/info/ld_region_get>`_.

        Parameters
        ----------
        population_name : str
            Population for which to compute LD. Use GET
            /info/variation/populations/:species?filter=LD to retrieve a list
            of all populations with LD data. Example values:
            '1000GENOMES:phase_3:KHV'.
        region : str
            Query region. A maximum region size of 500 kb is allowed. If the
            query region overlaps the MHC region only a maximum region size of
            10 kb is allowed. Example values: '6:25837556..25843455'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        d_prime : float or NoneType, optional, default: NoneType
            Measure of LD. If D' is provided only return pairs of variants
            whose D' value is equal to or greater than the value provided.
            Example values: '1.0'.
        r2 : float or NoneType, optional, default: NoneType
            Measure of LD. If r-squared is provided only return pairs of
            variants whose r-squared value is equal to or greater than the
            value provided. Example values: '0.85'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ld/{species}/region/{region}/{population_name}'.format(
            species=species,
            region=region,
            population_name=population_name
        )

        return self.rest_query(
            method_type,
            endpoint,
            d_prime=d_prime,
            r2=r2
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class LinkageDisequilibrium(brc._BaseRestClient, _LinkageDisequilibrium):
    """Perform LinkageDisequilibrium queries
    """
    pass
